# mypy: ignore-errors
# flake8: noqa

# This file is autogenerated by /metadata-ingestion/scripts/avro_codegen.py
# Do not modify manually!

# pylint: skip-file
# fmt: off
# isort: skip_file
from .....schema_classes import DocumentContentsClass
from .....schema_classes import DocumentInfoClass
from .....schema_classes import DocumentSettingsClass
from .....schema_classes import DocumentSourceClass
from .....schema_classes import DocumentSourceTypeClass
from .....schema_classes import DocumentStateClass
from .....schema_classes import DocumentStatusClass
from .....schema_classes import ParentDocumentClass
from .....schema_classes import RelatedAssetClass
from .....schema_classes import RelatedDocumentClass


DocumentContents = DocumentContentsClass
DocumentInfo = DocumentInfoClass
DocumentSettings = DocumentSettingsClass
DocumentSource = DocumentSourceClass
DocumentSourceType = DocumentSourceTypeClass
DocumentState = DocumentStateClass
DocumentStatus = DocumentStatusClass
ParentDocument = ParentDocumentClass
RelatedAsset = RelatedAssetClass
RelatedDocument = RelatedDocumentClass

# fmt: on
