# mypy: ignore-errors
# flake8: noqa

# This file is autogenerated by /metadata-ingestion/scripts/avro_codegen.py
# Do not modify manually!

# pylint: skip-file
# fmt: off
# isort: skip_file
from .....schema_classes import SchemaFieldAliasesClass
from .....schema_classes import SchemaFieldInfoClass


SchemaFieldAliases = SchemaFieldAliasesClass
SchemaFieldInfo = SchemaFieldInfoClass

# fmt: on
