# mypy: ignore-errors
# flake8: noqa

# This file is autogenerated by /metadata-ingestion/scripts/avro_codegen.py
# Do not modify manually!

# pylint: skip-file
# fmt: off
# isort: skip_file
from .....schema_classes import DataHubActorFilterClass
from .....schema_classes import DataHubPolicyInfoClass
from .....schema_classes import DataHubResourceFilterClass
from .....schema_classes import DataHubRoleInfoClass
from .....schema_classes import PolicyMatchConditionClass
from .....schema_classes import PolicyMatchCriterionClass
from .....schema_classes import PolicyMatchFilterClass


DataHubActorFilter = DataHubActorFilterClass
DataHubPolicyInfo = DataHubPolicyInfoClass
DataHubResourceFilter = DataHubResourceFilterClass
DataHubRoleInfo = DataHubRoleInfoClass
PolicyMatchCondition = PolicyMatchConditionClass
PolicyMatchCriterion = PolicyMatchCriterionClass
PolicyMatchFilter = PolicyMatchFilterClass

# fmt: on
